' //////////////////////////////////////////////////////////////////////
' TITLE: EPM_Defnder_Sound_Engine.spin
'
' DESCRIPTION:
'
' AUTHOR:  Eric Moyer, Copyright (C)2007
'
'
'  Rev  Date      Description
'  ---  --------  ---------------------------------------------------
'  001  11-17-07  Initial Creation
'  003  11-18-07  First working audio.
'  006  11-18-07  Increase edge rate on square waves.       
'  009  11-18-07  Add multi sound support.  Add play and stop controls.
' //////////////////////////////////////////////////////////////////////
CON

  CLOCK_RATE             = 80_000_000
  AUDIO_SAMPLE_RATE      = 22_000
  
  '------------------------------------
  'SOUND EFFECT IDs
  '------------------------------------

  SOUND__SINE            = %00000000_00000000_00000000_00000001 'Pure sine wave for testing
  SOUND__LEVEL_DONE      = %00000000_00000000_00000000_00000010 'Level Done Sound
  SOUND__FIRE            = %00000000_00000000_00000000_00000100 'Player shot fire
  SOUND__THRUST          = %00000000_00000000_00000000_00001000 'Player thrust
  SOUND__LANDER_DIE      = %00000000_00000000_00000000_00010000 'Lander death 

  PIN__AUDIO             = %00000000_00000000_00000000_10000000    

OBJ
  'dbg       :       "PasDebug"                       '<---- Add for Debugger
  
VAR


PUB start (sfx_control_block_p)

  cognew(@sfx_entry,sfx_control_block_p)
'  dbg.start(31,30,@sfx_entry)                 '<---- Add for Debugger   

  
'///////////////////////////////////////////////////////////////////////
' DATA SECTION /////////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

DAT
                        
'------------------------------------
'Entry
'------------------------------------
                        org        0
                        
sfx_entry

'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
'   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
'   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  -------------------------------------------------------------- 

                        'get pointers to parameters in the Effect Control block
                        mov     r1,PAR             'Get pointer to Effect Control block
                        mov     p_sfx_start_flags, r1
                        add     r1, #4
                        mov     p_sfx_stop_flags, r1

'------------------------------------
'Effect processing loop
'------------------------------------

                        '------------------------------------
                        'Init 
                        '------------------------------------
                        mov     sfx_playing_mask, #0
                        mov     wave_out, #0

                        'Configure audio pin
                        andn    outa,_PIN__AUDIO       ' Set audio pin LOW          "0"
                        or      dira,_PIN__AUDIO       ' Set audio pin to OUTPUT
                        mov     ctra, CTRAVAL
                        mov     frqa, LONG_HALF
                        mov     phsa, #$ff

                        'DEBUG LEG
                        or      dira, #1
                        

                        mov     start_request_previous, #0
                        mov     start_request_current, #0
                        mov     start_request_actions, #0
                        mov     sfx_in_progress, #0
:restart                        
                        mov     next_sample_start_time, cnt
                        add     next_sample_start_time, SAMPLE_INTERVAL_TICKS

                        mov     previous_pulse_level, #0


:sample_sync            'wait for next audio sample start time
                        waitcnt next_sample_start_time, SAMPLE_INTERVAL_TICKS

                        '------------------------------------
                        'Process SFX start requests
                        '------------------------------------

                        'Determine if any SFX's have been requested to start
                        rdlong  start_request_current, p_sfx_start_flags       'get current flags
                        mov     start_request_actions, start_request_current   'valid starts are any newly set bits
                        andn    start_request_actions, start_request_previous
                        mov     start_request_previous, start_request_current 'save current as previoius

                        '------------------------------------
                        'START: Level Done (0x0002)
                        '------------------------------------
                        mov     sfx_id_mask,#SOUND__LEVEL_DONE
                        test    start_request_actions, sfx_id_mask wz
                  if_z  jmp     #:sfx_start_fire


                        'Configure leveldone pulse
                        mov     pulse1_init, LONG_HALF
                        sub     pulse1_init, PULSE1_INIT_START
                        mov     pulse1_base_init, PULSE1_INIT_START
                        mov     pulse1_end, LONG_HALF
                        add     pulse1_end, PULSE1_END_START
                        
                        mov     pulse2_init, LONG_HALF
                        sub     pulse2_init, PULSE2_INIT_START
                        mov     pulse2_end, LONG_HALF
                        add     pulse2_end, PULSE2_END_START

                        mov     pulse1_val, pulse1_init
                        mov     pulse2_val, pulse2_init

                        'Indicate that the sound is in progress  
                        or     sfx_in_progress, sfx_id_mask
                        
                        
                        '------------------------------------
                        'START: Fire (0x0004)
                        '------------------------------------
:sfx_start_fire         shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz
                  if_z  jmp     #:sfx_start_thrust

                        mov     FIRE_phase_ticks, FIRE_PHASE_TICK_INIT
                        mov     FIRE_chunk_width, #1
                        mov     FIRE_pulse_ticks, #1
                        mov     FIRE_pulse_sense, #1

                        'Indicate that the sound is in progress
                        or     sfx_in_progress, sfx_id_mask 

                        '------------------------------------
                        'START: Thrust (0x0008)
                        '------------------------------------
:sfx_start_thrust       shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz
                  if_z  jmp     #:sfx_start_lander_die

                        '------------------------------------
                        'START: Lander Die (0x0010)
                        '------------------------------------
:sfx_start_lander_die   shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz
                  if_z  jmp     #:DONE_start

                        mov     LDIE_angle, #0
                        mov     LDIE_angle_step, LDIE_ANGLE_STEP_INIT
                        mov     LDIE_scale, #4

                        'Indicate that the sound is in progress
                        or     sfx_in_progress, sfx_id_mask 

:DONE_start

                        '================================================================================
                        'PLAY: Level Done (0x0002)
                        '================================================================================  
                        mov     sfx_id_mask,#SOUND__LEVEL_DONE
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:sfx_play_fire

                        
                        'Compute pulse waveforms
                        add     pulse1_val, PULSE_STEP              'increment pulse counter  
                        cmp     pulse1_end, pulse1_val   wc         'if pulse end reached
        if_nc           jmp     #:pulse2                            '{
                        add     pulse1_init, #2                     '   reduce low time by 1 sample
                        mov     pulse1_val, pulse1_init             '   adopt new pulse timing
                        
                        cmp     pulse1_init, LONG_HALF wc           '   if low time has reached zero
        if_c            jmp     #:pulse2                            '   {
                        sub     pulse1_end, PULSE_MOD_STEP          '      reduce high time by 1 sample
                        add     pulse1_base_init, PULSE_MOD_STEP
                        mov     pulse1_init, LONG_HALF              '      set low time back to starting width
                        sub     pulse1_init, pulse1_base_init

                        mov     pulse2_init, LONG_HALF
                        sub     pulse2_init, PULSE2_INIT_START
                                         '
                        cmp     pulse1_end, LONG_HALF   wc          '      if high time reached zero
                                                                    '      {
        if_c            andn    sfx_in_progress, sfx_id_mask        '         clear "in progress" mask (i.e. done)
                                                                    '      }      
                                                                    '   }
                                                                    '} 
:pulse2                 add     pulse2_val, PULSE_STEP              'increment pulse counter
                        cmp     pulse1_end, pulse2_val    wc        'if pulse end reached   
        if_c            mov     pulse2_val, pulse2_init             '   reload pulse 2



                        '------------------------------------
                        'Determine output waveform state
                        '------------------------------------
                        
                        'Output high if PULSE_BIT is high in either pulse generator
                        mov     r1, #00

                        cmp     pulse2_val, PULSE_BIT wc
              if_nc     mov     r1,#$ff
         
                        cmp     pulse1_val, PULSE_BIT wc
              if_nc     mov     r1,#$ff
                        
    
                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        mov     r2, r1
                        shl     r2, #21
                        add     r2, LONG_HALF


                        'Add edge emphasis if transitioning
                        cmp     previous_pulse_level, #0  wz
                  if_nz jmp     #:check_falling
                        cmp     r1, #$ff  wz
                   if_z add     r2, EDGE_EMPHASIS_INIT     'add endge emphasis
                        jmp     #:do_pwm

:check_falling          cmp     r1, #0   wz
                   if_z sub     r2, EDGE_EMPHASIS_INIT     'subtract edge emphasis

:do_pwm

                        mov     frqa, r2
                        mov     previous_pulse_level, r1   'remember old pulse  


                        '================================================================================
                        'PLAY: Fire (0x0004)
                        '================================================================================
:sfx_play_fire  
                        shl     sfx_id_mask, #1
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:sfx_play_thrust


                        sub     FIRE_phase_ticks, #1 wc
                  if_nc jmp     #:skip2
                        mov     FIRE_phase_ticks, FIRE_PHASE_TICK_INIT
                        add     FIRE_chunk_width, #1
                        cmp     FIRE_chunk_width, #60 wz
                  if_z  andn    sfx_in_progress, sfx_id_mask 
:skip2                            
                        sub     FIRE_pulse_ticks, #1  wz
                  if_nz jmp     #:skip3
                        add     random, RANDOM_DELTA
                        mov     x, random
                        shr     x, #29
                        and     x, #7
                        add     x, #1
                        mov     y, FIRE_chunk_width
                        call    #mult
                        mov     FIRE_pulse_ticks, y           
                        xor     FIRE_pulse_sense, #1

:skip3
                        'Output high if PULSE_BIT is high in either pulse generator
                        mov     r2, #00

                        cmp     FIRE_pulse_sense, #1 wz
                 if_z   mov     r2,#$ff

                        shl     r2, #21
                        add     r2, LONG_HALF
                        mov     frqa, r2  

:sfx_play_thrust
                        '================================================================================
                        'PLAY: Thrust (0x0008)
                        '================================================================================  
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:sfx_play_lander_die


:sfx_play_lander_die
                        '================================================================================
                        'PLAY: Lander Die (0x0010)
                        '================================================================================  
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:DONE_play

                        
                        '------------------------------------
                        'Compute waveform
                        '------------------------------------

                        'Get sine of current angle
                        mov     r1, LDIE_angle_step
                        shl     r1,#3
                        add     LDIE_angle, r1
                        mov     sin, LDIE_angle
                        shr     sin, #10
                        call    #getsin
                        add     sin, HALF_MAX_SIN
                        shr     sin, #9

                        'Scale by current scaling factor (4,3,2,1)
                        mov     x, sin
                        mov     y, LDIE_scale
                        call    #mult
                        mov     sin, y

                        'Clip
                        cmp     sin, #$1ff     wc
                 if_nc  mov     sin, #$1ff

                        'Decay the angle step
                        mov     x, LDIE_angle_step
                        mov     y, LDIE_ANGLE_MULTIPLY
                        call    #mult
                        shr     y, #15
                        mov     LDIE_angle_step, y

                        'When angle step has decayed update scaling factor
                        cmp     LDIE_angle_step, #$ff wz
                  if_nz jmp     #:skip1    
                        mov     LDIE_angle, #0
                        mov     LDIE_angle_step, LDIE_ANGLE_STEP_INIT
                        sub     LDIE_scale, #1 wz
                  if_z  andn    sfx_in_progress, sfx_id_mask
                        
:skip1
                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        mov     r2, sin
                        shl     r2, #21
                        add     r2, LONG_HALF

                        mov     frqa, r2
                        mov     previous_pulse_level, r1   'remember old pulse


                        '================================================================================
                        'Done Playing
                        '================================================================================  
:DONE_play  
                        
                        jmp     #:sample_sync


'------------------------------------ 
' Get sine/cosine
'
'       quadrant:    1            2            3            4
'          angle:    $0000..$07FF $0800..$0FFF $1000..$17FF $1800..$1FFF
'    table index:    $0000..$07FF $0800..$0001 $0000..$07FF $0800..$0001
'         mirror:    +offset      -offset      +offset      -offset
'           flip:    +sample      +sample      -sample      -sample
'
' on entry: sin[12..0] holds angle (0 to just under 360)
' on exit: sin holds signed value ranging from $0000FFFF ('1') to
' $FFFF0001 ('-1')
'------------------------------------ 
getcos                  add     sin,sin_90      'for cosine, add 90
getsin                  test    sin,sin_90 wc   'get quadrant 2|4 into c
                        test    sin,sin_180 wz  'get quadrant 3|4 into nz
                        negc    sin,sin         'if quadrant 2|4, negate offset
                        or      sin,sin_table   'or in sin table address >> 1
                        shl     sin,#1          'shift left to get final word address
                        rdword  sin,sin         'read word sample from $E000 to $F000
                        negnz   sin,sin         'if quadrant 3|4, negate sample
getsin_ret
getcos_ret              ret                     '39..54 clocks
                                                '(variance due to HUB sync on RDWORD)
sin_90                  long    $0800
sin_180                 long    $1000
sin_table               long    $E000 >> 1      'sine table base shifted right
sin                     long    0
HALF_MAX_SIN            long    $FFFF                               


'------------------------------------
'Multiply                                    
'------------------------------------
' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
'------------------------------------
mult                    shl x,#16               'get multiplicand into x[31..16]
                        mov t,#16               'ready for 16 multiplier bits
                        shr y,#1 wc             'get initial multiplier bit into c
:loop
                        if_c add y,x wc         'conditionally add multiplicand into product
                        rcr y,#1 wc             'get next multiplier bit into c.
                                                ' while shift product
                        djnz t,#:loop           'loop until done
mult_ret                ret                     'return with product in y[31..0]

'------------------------------------
'Initialized Data                                      
'------------------------------------
_PIN__AUDIO             long    PIN__AUDIO

SAMPLE_INTERVAL_TICKS   long   (CLOCK_RATE / AUDIO_SAMPLE_RATE)
'PULSE_BASELINE_TICKS    long   (CLOCK_RATE / AUDIO_SAMPLE_RATE)/5 
ZERO                    long   0
CTRAVAL                 long   %00110 << 26 + 7  'DUTY single-ended, APIN=7        '
LONG_HALF               long   $7FFF_FFFF
PULSE_BIT               long   $8000_0000
PULSE1_INIT_START       long   44 * 8
PULSE1_END_START        long   14 * 8
PULSE2_INIT_START       long   162 * 8
PULSE2_END_START        long   12 * 8
PULSE_MOD_STEP          long   10
PULSE_STEP              long   8

EDGE_EMPHASIS_INIT      long   $3fffffff
EDGE_EMPHASIS_DECAY     long   $3fffffff

BUZZ                    long   $05000000

LDIE_ANGLE_STEP_INIT    long   $0000693a
LDIE_ANGLE_MULTIPLY     long   $00007Fd9 

FIRE_PHASE_TICK_INIT    long   $00000300

RANDOM_DELTA            long   $9e3779b9
'------------------------------------
'Uninitialized Data
'------------------------------------
                          
r1                        res     1
r2                        res     1
sfx_id_mask               res     1

x                         res     1    ' for multiply
y                         res     1    ' for multiply
t                         res     1    ' for multiply

previous_pulse_level      res     1

start_request_previous    res     1
start_request_current     res     1
start_request_actions     res     1
sfx_in_progress           res     1
  
p_sfx_start_flags         res     1    
p_sfx_stop_flags          res     1
sfx_playing_mask          res     1

next_sample_start_time    res     1
pulse_end_time            res     1

wave_out                  res     1

pulse1_base_init          res     1
pulse1_init               res     1
pulse2_init               res     1
pulse1_end                res     1
pulse2_end                res     1
pulse1_val                res     1
pulse2_val                res     1

random                    res     1

'LADER DIE parameters
LDIE_angle                res     1
LDIE_angle_step           res     1
LDIE_scale                res     1

'FIRE parameters
FIRE_phase_ticks          res     1
FIRE_chunk_width          res     1
FIRE_pulse_ticks          res     1
FIRE_pulse_sense          res     1

                          fit                 