''//////////////////////////////////////////////////////////////////////////////
'' Defender Remake                                                        
'' AUTHOR: Steve Waddicor
''//////////////////////////////////////////////////////////////////////////////

OBJ
    g      : "sw_df_globals_002"
    tv     : "sw_df_tv_drv_005"
    gfx    : "sw_df_gfx_004"
    p_rend : "sw_df_gfx_renderer_007"
    s_rend : "sw_df_scanner_renderer_002"
    sfx    : "EPM_Defender_Sound_Engine_011.spin"

CON
    _clkmode = xtal1 + pll8x
    _xinfreq = 10_000_000 + 0000
    _stack = 76
    PLAYFIELD_RENDER_COGS = 4     
    SCANNER_RENDER_COGS = 1

CON
    SPRITE_XPOS                 = 0
    SPRITE_YPOS                 = 1
    SPRITE_INVISIBLE            = %1000_0000_0000_0000

CON
    FIRST_WAVE = 100
    HUMANOID_WAVE = 145
    SECOND_WAVE = 550

CON #0
    GAME_TITLE_INIT
    GAME_TITLE
    GAME_START
    GAME_DISPLAY_PLAYER_INIT
    GAME_DISPLAY_PLAYER
    GAME_LEVEL_1_INIT
    GAME_LEVEL_1
    GAME_OVER_INIT
    GAME_OVER

CON #0
    STATE_NOT_DEPLOYED
    STATE_APPEARING
    STATE_LANDER_MOVEMENT
    STATE_MUTANT_MOVEMENT
    STATE_BOMBER_MOVEMENT
    STATE_HUMANOID_MOVEMENT
    STATE_LANDER_TARGETED
    STATE_HUMANOID_TARGETED
    STATE_LANDER_ASCENDING
    STATE_HUMANOID_ASCENDING    
    STATE_DYING
    STATE_DEAD

CON
    #0
    TILE_PLAYER_RIGHT
    TILE_PLAYER_LEFT
    TILE_LANDER
    TILE_HUMANOID_LEFT
    TILE_HUMANOID_RIGHT
    TILE_MUTANT

VAR
    ' Renderer parameter block
    ' Each renderer copies these values to it's own cog_memory cache
    ' when it's got a little spare time, several times per frame
    ' so these values can be altered on the fly.
    long    rend_cog_number        'each cog gets a different number, starting with zero     write-only
    long    rend_playfield_cog_count         'total number of cogs                                     write-only
    long    rend_scanner_cog_count
    long    rend_scanline_req_adr  'a pointer to the request to put a line in the scan buffer   write-only
    long    rend_sprite_def_ptr    'pointer to tile definitions (longs)                      write-only
    long    rend_laser_def_ptr
    long    rend_palette_ptr       'pointer to colors (longs)                               write-only
    long    rend_scanline_buffer_adr 'hub memory where renderer puts scanline when requested write-only
    long    rend_terrain_ptr
    long    rend_side_scroll                         'position of screen left in pixels (wraps at 1024)
    long    rend_sprite_locs[g#NUMBER_OF_SPRITES]    '%i.YYYYYY_YYyyyyyy_XXXXXXXX_XXxxxxxx (negative = invisible)
    long    rend_sprite_tiles[g#NUMBER_OF_SPRITES]   '%EEEEEEEE_........_........_TTTTTTTT (E = explode count, T = tile number)
    long    rend_lasers[g#NUMBER_OF_LASER_SHOTS]     '%.......R_NNNNNNNN_XXXXXXXX_YYYYYYYY (N = state_count, Y = scan line, R= reverse)
    long    rend_palette[g#NUMBER_OF_COLORS]
                               
    'End of renderer parameter block

VAR
    ' TV/KB driver parameter block
    long    kb_tail        'key buffer tail        read/write      (20 contiguous longs)
    long    kb_head        'key buffer head        read-only
    long    kb_present     'keyboard present       read-only
    long    kb_states[8]   'key states (256 bits)  read-only
    long    kb_keys[8]     'key buffer (16 words)  read-only       (also used to pass initial parameters)
    long    kb_break       'break key code, flag   read/write
    long    tv_scanline_request
    long    tv_scanline_buffer
    long    tv_status
    ' End of TVTV/KB parameter block

VAR
    long  startkey
    long  selectkey
    long  debug

VAR
    byte    game_state
    byte    player_level[2]
    byte    player_lives[2]
    byte    game_player_up
    byte    game_continue_life
    long    mutant_color

    long    game_two_players
    long    player_score[2]
    long    hi_score
    long    gp
    long    game_state_count
    long    timer
    long    timer_state_count
    long    rnd

    ' New for Defender
    long    player_x_velocity
    long    player_y_velocity
    long    player_h_direction
    long    player_v_direction
    long    target_player_position
    long    player_position
    long    explode_count
    long    explode_state_count
    long    x_velocity[g#NUMBER_OF_SPRITES]
    long    y_velocity[g#NUMBER_OF_SPRITES]
    long    laser_state_count
    long    a_button_up
    long    frame_count
    long    color_count

    long    sprite_state[g#NUMBER_OF_SPRITES]
    long    target[g#NUMBER_OF_SPRITES]

    ' should be locals, but need to be words
    word    laser_left
    word    laser_right 

    'Sound control block
    long    sfx_start_flags
    long    sfx_stop_flags

PUB Main | i, time, game_counter, total_time, avg_time, t
    'Mark top of stack to check later
    long[$7FFC]:=$DEAD_BEEF
    'Start renderers
    rend_sprite_def_ptr := gfx.sprite_adr
    rend_laser_def_ptr := gfx.laser_adr
    rend_palette_ptr := 0
    rend_scanline_buffer_adr := @hub_scanline_buffer
    rend_scanline_req_adr := @scanline_req
    rend_playfield_cog_count := PLAYFIELD_RENDER_COGS
    rend_scanner_cog_count := SCANNER_RENDER_COGS
    rend_terrain_ptr := gfx.terrain_adr
    rend_palette[0]:=$00000002 'black
    rend_palette[1]:=$0000001b 'purple (ship wing)
    rend_palette[2]:=$0000005b 'red
    rend_palette[3]:=$00000006 'white
    rend_palette[4]:=$000000bc 'green
    rend_palette[5]:=$0000007e 'yellow
    rend_palette[6]:=$0000006d 'brown (terrain)
    rend_palette[7]:=$0000004c 'pink 
    rend_palette[8]:=$00000004 'grey
    rend_palette[9]:=$0000009d
    rend_palette[10]:=$000000ad
    rend_palette[11]:=$000000bd
    rend_palette[12]:=$000000cd
    rend_palette[13]:=$000000dd
    rend_palette[14]:=$0000000c 'mutant color cycle
    rend_palette[15]:=$000000fd 'laser color cycle
    mutant_color := rend_palette[14]
    
    SpriteClear

    'Start the sound effects engine
    sfx.start(@sfx_start_flags)

    repeat i from 0 to constant(PLAYFIELD_RENDER_COGS-1)
        rend_cog_number := i
        p_rend.start(@rend_cog_number)
        'Wait to allow this cog to get the cog number param from the block before changing it.
        repeat 1000

    repeat i from 0 to constant(SCANNER_RENDER_COGS-1)
        rend_cog_number := PLAYFIELD_RENDER_COGS+i
        s_rend.start(@rend_cog_number)
        'Wait to allow this cog to get the cog number param from the block before changing it.
        repeat 1000

    tv_scanline_request := @scanline_req
    tv_scanline_buffer := @hub_scanline_buffer
    tv.start(@tv_scanline_request)

    'time:=CNT
    rnd:=$439845f0
    hi_score:=7650

    'Main Game Loop
    repeat
        'Check stack has not been overrun
        if long[$7FFC] <> $DEAD_BEEF
            DIRA[0]:=1
            OUTA[0]:=1
        'TV will let up know when there is a vsync
        tv_status := 0
        gp := NES_Read_Gamepad

        ' Clear sound API.  Start and stops happen on low to high transition.  So start low for eaxh frame.
        sfx_start_flags := 0
        sfx_stop_flags := 0

        ' Color cycling
        color_count:=frame_count++ & $7
        if color_count==0
            rend_palette[15]+=$10
            rend_palette[15]&=$FF
            rend_palette[14]:=$02
        elseif color_count==3
            rend_palette[15]+=$10
            rend_palette[15]&=$FF
            mutant_color+=$40
            mutant_color&=$FF
            rend_palette[14]:=mutant_color

        case game_state
            GAME_TITLE_INIT:
                SpriteClear
                game_state:=GAME_TITLE
            GAME_TITLE:
                if gp & GP_LEFT
                    game_two_players:=FALSE
                if gp & GP_RIGHT
                    game_two_players:=TRUE    
                'if gp & GP_START
                game_state:=GAME_START
            GAME_START:
                game_player_up := -game_two_players ' This is zero based.  Pretend player 2 was up last tyo make player 1 up this time.  
                game_state := GAME_LEVEL_1_INIT
            GAME_LEVEL_1_INIT:
                InitLevelCommon
                game_state := GAME_LEVEL_1
                rend_sprite_locs[0]:= $20_00_10_00
                rend_sprite_tiles[0]:= 0
                rnd := 419225
                player_h_direction:=1
                player_v_direction:=0
                repeat i from g#FIRST_HUMANOID to g#LAST_HUMANOID
                    sprite_state[i] := STATE_NOT_DEPLOYED
                repeat i from g#FIRST_LANDER to g#LAST_LANDER
                    sprite_state[i] := STATE_NOT_DEPLOYED
                repeat i from g#FIRST_BOMBER to g#LAST_BOMBER
                    sprite_state[i] := STATE_NOT_DEPLOYED
                repeat i from 0 to constant(g#NUMBER_OF_LASER_SHOTS-1)
                    rend_lasers[i] := $ff
                sfx_start_flags := sfx#SOUND__LEVEL_DONE
            GAME_LEVEL_1:
                PlayerLogic
                LaserLogic
                AlienLogic
            GAME_OVER_INIT:
                game_state:=GAME_OVER
            GAME_OVER:
                if game_state_count++ == 192
                    game_state:=GAME_TITLE_INIT

        repeat until tv_status>0   
       'repeat 100000

PRI SpriteClear | i
    repeat i from 0 to constant(g#NUMBER_OF_SPRITES-1)
        rend_sprite_locs[i]:= $8000_0000
        rend_sprite_tiles[i]:= 0

PRI InitLevelCommon
    SpriteClear

PRI PlayerLogic | i,advance,laser,player_base_position

    if (gp & GP_LEFT)
        player_h_direction := -1
        rend_sprite_tiles[0] := 1
    elseif (gp & GP_RIGHT)
        player_h_direction := 1
        rend_sprite_tiles[0] := 0

    if (gp & (GP_B | GP_LEFT | GP_RIGHT))
        player_x_velocity+=player_h_direction*4
    if player_x_velocity>0
        player_x_velocity-=2
    elseif player_x_velocity<0
        player_x_velocity+=2


 
    if (gp & GP_UP)
        if player_v_direction<>-1
            player_v_direction:=-1
            player_y_velocity:=0
        player_y_velocity-=6
    elseif (gp & GP_DOWN)
        if player_v_direction<>1
            player_v_direction:=1
            player_y_velocity:=0
        player_y_velocity+=6
    else
        player_y_velocity:=0

    player_x_velocity <#= 220
    player_x_velocity #>= -220
    player_y_velocity <#= 128
    player_y_velocity #>= -128
                      
    rend_sprite_locs.word[SPRITE_XPOS]+=player_x_velocity
    rend_sprite_locs.word[SPRITE_YPOS]+=player_y_velocity
    rend_sprite_locs.word[SPRITE_YPOS]<#=constant(189<<6)
    rend_sprite_locs.word[SPRITE_YPOS]#>=constant(4<<6)
'    rend_sprite_locs.word[SPRITE_YPOS]:=WrapToY(rend_sprite_locs.word[SPRITE_YPOS])
    
    advance := ((player_x_velocity/6)-8)#>0
    'target_player_position
    player_base_position := 84-player_h_direction*48
    target_player_position := player_base_position+advance
    if player_position>target_player_position
        player_position-=1
    elseif player_position<target_player_position
        player_position+=1
    rend_side_scroll := ((rend_sprite_locs.word[SPRITE_XPOS]>>6)-player_position) & %1111111111

    if (gp & GP_A)
        if a_button_up
            laser := ((rend_sprite_locs.word[SPRITE_YPOS]>>6)+1) ' y pos
            laser += (((||(player_position-player_base_position)))<#29) << 16 'tile
            if player_h_direction == 1
                laser|=$2800
            else
                laser|=$100_7f00
            rend_lasers[laser_state_count] :=laser
            laser_state_count++
            if laser_state_count==g#NUMBER_OF_LASER_SHOTS
                laser_state_count:=0
            sfx_start_flags |= sfx#SOUND__FIRE
        a_button_up := FALSE
    else
        a_button_up := TRUE

PRI AlienLogic | i, this_alien, this_humanoid, explode, h, x, y, terrain, distance, deltax, deltay, filter
    case frame_count
        FIRST_WAVE:
            repeat i from g#FIRST_LANDER to constant(g#MIDDLE_LANDER-1)
                rend_sprite_tiles[i]:= (45 << 16) + 2
                rend_sprite_locs[i]:= (19<<22) + (?rnd&$FFC0)
                sprite_state[i] := STATE_APPEARING
                x_velocity[i] := (?rnd&$F)+16
                if ?rnd&1
                    x_velocity[i] := -x_velocity[i]    
            repeat i from g#FIRST_BOMBER to g#LAST_BOMBER
                this_alien:= @rend_sprite_locs[i]
                rend_sprite_tiles[i]:= (45 << 16) + 6
                rend_sprite_locs[i]:= WrapToY((?rnd&$FF)<<6)<<16 + (?rnd&$1FC0)
                word[this_alien][SPRITE_XPOS]:=?rnd&$1FC0
                word[this_alien][SPRITE_YPOS]:=WrapToY(  (Sin(word[this_alien][SPRITE_XPOS])>>(i-g#FIRST_BOMBER)   )+(96<<6))
                sprite_state[i] := STATE_APPEARING
                x_velocity[i] := 32
                if ?rnd&1
                    x_velocity[i] := -x_velocity[i]
            repeat i from g#FIRST_HUMANOID to g#LAST_HUMANOID
                rend_sprite_tiles[i]:= 4
                rend_sprite_locs[i]:= (180<<22) + (?rnd&$FFFF)
                sprite_state[i] := STATE_HUMANOID_MOVEMENT
                x_velocity[i] := (?rnd&%11)+1
                if ?rnd&1
                    x_velocity[i] := -x_velocity[i]    
                    rend_sprite_tiles[i]:= 3
        SECOND_WAVE:
            repeat i from g#MIDDLE_LANDER to g#LAST_LANDER
                rend_sprite_tiles[i]:= (45 << 16) + 2
                rend_sprite_locs[i]:= (19<<22) + (?rnd&$FFC0)
                sprite_state[i] := STATE_APPEARING
                x_velocity[i] := (?rnd&$F)+16
                if ?rnd&1
                    x_velocity[i] := -x_velocity[i]    

    repeat i from g#FIRST_HUMANOID to g#LAST_BOMBER
        this_alien:= @rend_sprite_locs[i]
        case sprite_state[i]
            STATE_APPEARING:
                if frame_count&1
                    explode := rend_sprite_tiles[i]>>16
                    if explode==0
                        if i=<g#LAST_LANDER
                            sprite_state[i] := STATE_LANDER_MOVEMENT
                        elseif i=<g#LAST_BOMBER
                            sprite_state[i] := STATE_BOMBER_MOVEMENT
                    else
                        explode--
                        rend_sprite_tiles[i] &= $FFFF
                        rend_sprite_tiles[i] |= (explode<<16)

            STATE_LANDER_MOVEMENT:
                repeat h from g#FIRST_HUMANOID to g#LAST_HUMANOID
                    'Possibly the lander will target the humanoid
                    if sprite_state[h]==STATE_HUMANOID_MOVEMENT
                        this_humanoid := @rend_sprite_locs[h]
                        if word[this_humanoid][sprite_XPOS]&$FFC0 == word[this_alien][sprite_XPOS]&$FFC0
                            sprite_state[i] := STATE_LANDER_TARGETED
                            sprite_state[h] := STATE_HUMANOID_TARGETED
                            target[i] := h
                 
                word[this_alien][SPRITE_XPOS]+=x_velocity[i]
                 
                x := word[this_alien][SPRITE_XPOS]>>6
                y := word[this_alien][SPRITE_YPOS]>>6
                terrain := byte[rend_terrain_ptr][x]
                if (terrain-y) > 40+i
                    ' head down
                    y_velocity[i]:=32
                elseif (terrain-y) < 16+i
                    'head up
                    y_velocity[i]:=-16
                else
                    y_velocity[i]:=0
                word[this_alien][SPRITE_YPOS]+=y_velocity[i]

            STATE_HUMANOID_MOVEMENT:
                'filter := (frame_count&$F) + constant(g#FIRST_HUMANOID+1)
                'if i == filter
                    word[this_alien][SPRITE_XPOS]+=x_velocity[i]
                     
                    x := word[this_alien][SPRITE_XPOS]>>6
                    y := word[this_alien][SPRITE_YPOS]>>6
                    terrain := byte[rend_terrain_ptr][x]
                    if (y-terrain) <10
                        ' head down
                        y_velocity[i]:=32
                    elseif (y-terrain) > 12
                        'head up
                        y_velocity[i]:=-16
                    else
                        y_velocity[i]:=0
                    word[this_alien][SPRITE_YPOS]+=y_velocity[i]

            STATE_BOMBER_MOVEMENT:
                word[this_alien][SPRITE_XPOS]+=x_velocity[i]
                word[this_alien][SPRITE_YPOS]:=WrapToY(  (Sin(word[this_alien][SPRITE_XPOS])>>(i-g#FIRST_BOMBER)   )+(96<<6))

            STATE_LANDER_TARGETED:
                this_humanoid := @rend_sprite_locs[target[i]]
                distance := (word[this_humanoid][SPRITE_YPOS]>>6)-(word[this_alien][SPRITE_YPOS]>>6)
                if distance<13
                    sprite_state[i] := STATE_LANDER_ASCENDING
                    sprite_state[target[i]] := STATE_HUMANOID_ASCENDING
                else
                    word[this_alien][SPRITE_YPOS]+=16

            STATE_LANDER_ASCENDING:
                this_humanoid := @rend_sprite_locs[target[i]]
                word[this_alien][SPRITE_YPOS]-=32
                word[this_humanoid][SPRITE_YPOS]-=32
                if word[this_alien][SPRITE_YPOS] == constant(8<<6)
                    sprite_state[target[i]]:=STATE_DEAD
                    long[this_humanoid]:=-1
                    sprite_state[i]:=STATE_MUTANT_MOVEMENT
                    rend_sprite_tiles[i]:= 5

            STATE_MUTANT_MOVEMENT:
                ' delta x is shifted left to wrap around at 1024 pixel landsacape width.  i.e. -512 to +511
                deltax := (word[this_alien][SPRITE_XPOS]-rend_sprite_locs.word[SPRITE_XPOS])
                deltay := (word[this_alien][SPRITE_YPOS]-rend_sprite_locs.word[SPRITE_YPOS])
                if deltax > 80
                    x_velocity[i] := -48
                    CalcDistantMutantY(i,deltay)    
                elseif deltax < -80
                    x_velocity[i] := 48
                    CalcDistantMutantY(i,deltay)    
                else
                    CalcCloseMutantY(i,deltay)
                word[this_alien][SPRITE_XPOS]+=x_velocity[i]
                word[this_alien][SPRITE_YPOS]+=y_velocity[i]

            STATE_DYING:
                if frame_count&1
                    explode := rend_sprite_tiles[i]>>16
                    if explode==45
                        sprite_state[i] := STATE_DEAD
                        rend_sprite_locs[i]:=-1
                    else
                        explode++
                        rend_sprite_tiles[i] &= $FFFF
                        rend_sprite_tiles[i] |= (explode<<16)

            
            
PRI CalcDistantMutantY(i,deltay)
    if deltay<(8<<6) and deltay>(-8<<6)
        CalcCloseMutantY(i,deltay)
        y_velocity[i]:= -y_velocity[i]
    else
        if frame_count&3
            y_velocity[i]:=0
        elseif ?rnd&1
            y_velocity[i]:=64
        else
            y_velocity[i]:=-64

        
      
PRI CalcCloseMutantY(i,deltay) | r
    r := ?rnd&$F
    if r=<0
        y_velocity[i]:=-64
    elseif r=<7
        y_velocity[i]:=0
    elseif r=<13
        y_velocity[i]:=64
    else
        y_velocity[i]:=128
    if deltay>0
        y_velocity[i]:= -y_velocity[i]

PRI WrapToY(y)
    y:=~~y
    if y<6
        y+=constant(192<<6)
    elseif y>constant(192<<6)
        y-=constant(192<<6)
    return y

PRI Sin(angle) : y
''Get sine of angle (0-$1FFF)
    angle>>=3
    y := angle << 1 & $FFE
    if angle & $800
       y := word[$F000 - y]
    else
       y := word[$E000 + y]
    if angle & $1000
       -y

    y<<=15
    y~>=17

           
PRI LaserLogic | i,e,this_laser, this_enemy,tile, ly, ex,ey, reverse, current_ex, current_dead_alien, ex_extended 
    repeat i from 0 to constant(g#NUMBER_OF_LASER_SHOTS-1)
        this_laser:=@rend_lasers[i]                                                                    
        ly := byte[this_laser][0]
        tile := byte[this_laser][2]
        reverse := byte[this_laser][3]
        if tile==g#NUMBER_OF_LASER_FRAMES
            byte[this_laser][0] := $FF
        else
            byte[this_laser][2] := ++tile
            if not reverse
                laser_right:=(constant($28+8) +tile*4 +rend_side_scroll)
                laser_left:=($28 + tile + rend_side_scroll)
                current_ex := laser_right
            else
                laser_left:=(constant($7f-8) -tile*4 +rend_side_scroll)
                laser_right:=($7f - tile + rend_side_scroll)
                current_ex := laser_left            
            current_dead_alien := -1
            if byte[this_laser][0] <> $FF
            repeat e from g#FIRST_HUMANOID to g#LAST_BOMBER
                if sprite_state[e]>STATE_APPEARING and  sprite_state[e]<STATE_DYING
                    this_enemy:=@rend_sprite_locs[e]
                    ey := word[this_enemy][SPRITE_YPOS]>>6
                    if (ly =< ey+3) and (ly=>ey-4)' and (sprite_state[e] => STATE_GENERAL_MOVEMENT) and  (sprite_state[e] =< STATE_ASCENDING)
                        ex := word[this_enemy][SPRITE_XPOS]>>6
                        if not reverse
                            ex_extended := ex+g#TERRAIN_WIDTH
                            if  (ex<current_ex)                
                                if ((ex>laser_left) and (ex<laser_right)) or ((ex_extended>laser_left) and (ex_extended<laser_right))
                                    current_ex:=ex
                                    current_dead_alien:=e
                        else
                            ex_extended := ex-g#TERRAIN_WIDTH
                            if  (ex>current_ex)                
                                if ((ex>laser_left) and (ex<laser_right)) or ((ex_extended>laser_left) and (ex_extended<laser_right))
                                    current_ex:=ex
                                    current_dead_alien:=e
             
            if current_dead_alien>0
                sprite_state[current_dead_alien]:=STATE_DYING
                byte[this_laser][0] := $FF
                sfx_start_flags |= sfx#SOUND__LANDER_DIE
                DIRA[0]:=1
                OUTA[0]:=1
        this_laser += 4
 
{
        laser_def
    long %11101111000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %01101110111100000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00101110110111110000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00001110110110111111000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000110110110110111111100000000,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000110110110110110111111110000,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000010100110110110110111111111,%00000000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000100110110110110111011111,%11110000000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000100110100110110111010111,%11111111000000000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000110100110110111010110,%11111111111100000000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000100100110110111010110,%11011111111111110000000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000100100110110011010110,%11011011111111111111000000000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000000100110110011010110,%11011011011111111111111100000000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000000100110010011010110,%11011011011011111111111111110000,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000000100110010011010110,%11011011011010111111111111111111,%00000000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000000000110010011010110,%11011011011010111011111111111111,%11110000000000000000000000000000,%00000000000000000000000000000000
    long %00000000000000000110010001010110,%11011011011010111011011111111111,%11111111000000000000000000000000,%00000000000000000000000000000000
    long %00000000000000000110010001010110,%11011011011010111011011011111111,%11111111111100000000000000000000,%00000000000000000000000000000000
    long %00000000000000000010010001010110,%11011011011010111011011011011111,%11111111111111110000000000000000,%00000000000000000000000000000000
    long %00000000000000000000010001010110,%11011011011010111011011011011011,%11111111111111111111000000000000,%00000000000000000000000000000000
    long %00000000000000000000010001010110,%11011011011010111011011011011010,%11111111111111111111111100000000,%00000000000000000000000000000000
    long %00000000000000000000010001010110,%11011011011010111011011011011010,%11101111111111111111111111110000,%00000000000000000000000000000000
    long %00000000000000000000000001010110,%11011011011010111011011011011010,%11101101111111111111111111111111,%00000000000000000000000000000000
    long %00000000000000000000000001010110,%11011011011010111011011011011010,%11101101101111111111111111111111,%11110000000000000000000000000000
    long %00000000000000000000000001010110,%11011011011010111011011011011010,%11101101101011111111111111111111,%11111111000000000000000000000000
    long %00000000000000000000000001010110,%11011011011010111011011011011010,%11101101101011101111111111111111,%11111111111100000000000000000000
    long %00000000000000000000000000010110,%11011011011010111011011011011010,%11101101101011101101111111111111,%11111111111111110000000000000000
    long %00000000000000000000000000010110,%11011011011010111011011011011010,%11101101101011101101101111111111,%11111111111111111111000000000000
    long %00000000000000000000000000000110,%11011011011010111011011011011010,%11101101101011101101101101111111,%11111111111111111111111100000000
    long %00000000000000000000000000000110,%11011011011010111011011011011010,%11101101101011101101101101101111,%11111111111111111111111111110000
    long %00000000000000000000000000000010,%11011011011010111011011011011010,%11101101101011101101101101101011,%11111111111111111111111111111111
    long %00000000000000000000000000000000,%11011011011010111011011011011010,%11101101101011101101101101101010,%11111111111111111111111111111111
}
    
PRI Rotate(item_ptr,increment,lower,upper)
    ' Rotate a byte value through a range of values.  Useful for sprite animation.
    ' byte_item = pointer to a variable to ,reset,lower,upper)
    ' increment = value to add or subtract
    ' lower     = lower bound
    ' upper     = upper bound
    long[item_ptr]+=increment
    if long[item_ptr]>upper
        long[item_ptr] := lower
    if long[item_ptr]<lower
        long[item_ptr] := upper
    return

CON
' //////////////////////////////////////////////////////////////////
' NES Game Paddle Read
' //////////////////////////////////////////////////////////////////
' reads both gamepads in parallel encodes 8-bits for each in format
' right game pad #1 [15..8] : left game pad #0 [7..0]

    GP_RIGHT  = %00000001
    GP_LEFT   = %00000010
    GP_DOWN   = %00000100
    GP_UP     = %00001000
    GP_START  = %00010000
    GP_SELECT = %00100000
    GP_B      = %01000000
    GP_A      = %10000000

PUB NES_Read_Gamepad : nes_bits        |  i

' set I/O ports to proper direction
' P3 = JOY_CLK      (4)
' P4 = JOY_SH/LDn   (5)
' P5 = JOY_DATAOUT0 (6)
' P6 = JOY_DATAOUT1 (7)
' NES Bit Encoding

' step 1: set I/Os
DIRA [3] := 1 ' output
DIRA [4] := 1 ' output
DIRA [5] := 0 ' input
DIRA [6] := 0 ' input

' step 2: set clock and latch to 0
OUTA [3] := 0 ' JOY_CLK = 0
OUTA [4] := 0 ' JOY_SH/LDn = 0
'Delay(1)

' step 3: set latch to 1
OUTA [4] := 1 ' JOY_SH/LDn = 1
'Delay(1)

' step 4: set latch to 0
OUTA [4] := 0 ' JOY_SH/LDn = 0

' step 5: read first bit of each game pad

' data is now ready to shift out
' first bit is ready
nes_bits := 0

' left controller
nes_bits := INA[5] | (INA[6] << 8)

' step 7: read next 7 bits
repeat i from 0 to 6
 OUTA [3] := 1 ' JOY_CLK = 1
 'Delay(1)
 OUTA [3] := 0 ' JOY_CLK = 0
 nes_bits := (nes_bits << 1)
 nes_bits := nes_bits | INA[5] | (INA[6] << 8)

 'Delay(1)
' invert bits to make positive logic
nes_bits := (!nes_bits & $FFFF)

' //////////////////////////////////////////////////////////////////
' End NES Game Paddle Read
' //////////////////////////////////////////////////////////////////

DAT
' There seems to be some pointer leakage here.  Extra long as a patch.
                        long    0
' Hub scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
hub_scanline_buffer
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7c_7c_7c_7c
colorbar30              long    $7d_7d_7d_7d
colorbar31              long    $7e_7e_7e_7e
colorbar32              long    $8b_8b_8b_8b
colorbar33              long    $8c_8c_8c_8c
colorbar34              long    $8d_8d_8d_8d
colorbar35              long    $8e_8e_8e_8e
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe
colorbar64              long    $0b_0b_0b_0b
colorbar65              long    $0c_0c_0c_0c
colorbar66              long    $0d_0d_0d_0d
colorbar67              long    $0e_0e_0e_0e
colorbar68              long    $1b_0b_1b_1b
colorbar69              long    $1c_0c_1c_1c
colorbar70              long    $1d_0d_1d_1d
colorbar71              long    $1e_0e_1e_1e
colorbar72              long    $2b_0b_2b_2b
colorbar73              long    $2c_2c_2c_2c
colorbar74              long    $2d_2d_2d_2d
colorbar75              long    $2e_2e_2e_2e
colorbar76              long    $3b_3b_3b_3b
colorbar77              long    $3c_3c_3c_3c
colorbar78              long    $3d_3d_3d_3d
colorbar79              long    $3e_3e_3e_3e
colorbar80              long    $4b_4b_4b_4b

' End of scanline buffer

scanline_req            long    0

 